import numpy as np
from astropy.coordinates import SkyCoord
from astroquery.ipac.ned import Ned
from dustmaps.planck import PlanckQuery
from pathlib import Path

class Reddening:
    def __init__(self):
        from dustmaps.config import config
        config.reset()
        
        import dustmaps.planck
        dustmaps.planck.fetch()
        
    
    def dered(self, image_data, error_data, galaxy_name, observatory, band, image_set):
        self.obj, self.obs, self.filt = galaxy_name, observatory, band
        self.filter_file = f'{self.obs}.{self.filt}.dat'
        self.ra, self.dec = Ned.query_object(self.obj)['RA', 'DEC'][0]
        coords = SkyCoord(self.ra, self.dec, unit='deg', frame='icrs')
        planck = PlanckQuery()
        self.ebv = planck(coords)
        wave, resp = self.get_resp_curve()
        # Check if filter is valid for CCM98 model
        if (max(wave) > 3.3 * 1e4) | (min(wave) < 9.1 * 1e2):
            print('Filter is not valid for CCM98 model')
            return 0
        
        A_mean = self.median_reddening(wave, resp)
        deredden_img = image_data * 10 ** (0.4 * A_mean)
        deredden_err = error_data * 10 ** (0.4 * A_mean)
        
        image_set.update_data(deredden_img, galaxy_name, observatory, band)
        image_set.update_error(deredden_err, galaxy_name, observatory, band)
        
                
    def get_resp_curve(self):
        import importlib.resources

        dat_path = importlib.resources.files("Spec7DT.reference.filter_curves").joinpath(f"{self.filter_file}")
        
        f_g = np.genfromtxt(dat_path, skip_header = 3 , delimiter = ' ' , dtype = float)
        wave = f_g[:, 0]
        resp = f_g[:, 1]
        
        mask = (resp != 0)
        
        return wave[mask], resp[mask]
    
    def reddening_ccm(self, wave, ebv=None, a_v=None, r_v=3.1, model='ccm89'):
        """
        Not used in FIREFLY
        Determines a CCM reddening curve.

        Parameters
        ----------
        wave: ~numpy.ndarray
            wavelength in Angstroms
        flux: ~numpy.ndarray
        ebv: float
            E(B-V) differential extinction; specify either this or a_v.
        a_v: float
            A(V) extinction; specify either this or ebv.
        r_v: float, optional
            defaults to standard Milky Way average of 3.1
        model: {'ccm89', 'gcc09'}, optional
            * 'ccm89' is the default Cardelli, Clayton, & Mathis (1989) [1]_, but
            does include the O'Donnell (1994) parameters to match IDL astrolib.
            * 'gcc09' is Gordon, Cartledge, & Clayton (2009) [2]_. This paper has
            incorrect parameters for the 2175A bump; not yet corrected here.

        Returns
        -------
        reddening_curve: ~numpy.ndarray
            Multiply to deredden flux, divide to redden.

        Notes
        -----
        Cardelli, Clayton, & Mathis (1989) [1]_ parameterization is used for all
        models. The default parameter values are from CCM except in the optical
        range, where the updated parameters of O'Donnell (1994) [3]_ are used
        (matching the Goddard IDL astrolib routine CCM_UNRED).

        The function is works between 910 A and 3.3 microns, although note the
        default ccm89 model is scientifically valid only at >1250 A.

        Model gcc09 uses the updated UV coefficients of Gordon, Cartledge, & Clayton
        (2009) [2]_, and is valid from 910 A to 3030 A. This function will use CCM89
        at longer wavelengths if GCC09 is selected, but note that the two do not
        connect perfectly smoothly. There is a small discontinuity at 3030 A. Note
        that GCC09 equations 14 and 15 apply to all x>5.9 (the GCC09 paper
        mistakenly states they do not apply at x>8; K. Gordon, priv. comm.).

        References
        ----------
        [1] Cardelli, J. A., Clayton, G. C., & Mathis, J. S. 1989, ApJ, 345, 245
        [2] Gordon, K. D., Cartledge, S., & Clayton, G. C. 2009, ApJ, 705, 1320
        [3] O'Donnell, J. E. 1994, ApJ, 422, 158O

        """

        import warnings

        model = model.lower()
        if model not in ['ccm89','gcc09']:
            raise ValueError('model must be ccm89 or gcc09')
        if (a_v is None) and (ebv is None):
            raise ValueError('Must specify either a_v or ebv')
        if (a_v is not None) and (ebv is not None):
            raise ValueError('Cannot specify both a_v and ebv')
        if a_v is not None:
            ebv = a_v / r_v

        if model == 'gcc09':
            raise ValueError('TEMPORARY: gcc09 currently does 2175A bump '+
                'incorrectly')

        x = 1e4 / wave      # inverse microns
        if any(x < 0.3) or any(x > 11):
            raise ValueError('ccm_dered valid only for wavelengths from 910 A to '+
                '3.3 microns')
        if any(x > 8) and (model == 'ccm89'):
            warnings.warn('CCM89 should not be used below 1250 A.')
        #    if any(x < 3.3) and any(x > 3.3) and (model == 'gcc09'):
        #        warnings.warn('GCC09 has a discontinuity at 3030 A.')

        a = np.zeros(x.size)
        b = np.zeros(x.size)

        # NIR
        valid = (0.3 <= x) & (x < 1.1)
        a[valid] = 0.574 * x[valid]**1.61
        b[valid] = -0.527 * x[valid]**1.61

        # optical, using O'Donnell (1994) values
        valid = (1.1 <= x) & (x < 3.3)
        y = x[valid] - 1.82
        coef_a = np.array([-0.505, 1.647, -0.827, -1.718, 1.137, 0.701, -0.609,
            0.104, 1.])
        coef_b = np.array([3.347, -10.805, 5.491, 11.102, -7.985, -3.989, 2.908,
            1.952, 0.])
        a[valid] = np.polyval(coef_a,y)
        b[valid] = np.polyval(coef_b,y)

        # UV
        valid = (3.3 <= x) & (x < 8)
        y = x[valid]
        f_a = np.zeros(y.size)
        f_b = np.zeros(y.size)
        select = (y >= 5.9)
        yselect = y[select] - 5.9

        f_a[select] = -0.04473 * yselect**2 - 0.009779 * yselect**3
        f_b[select] = 0.2130 * yselect**2 + 0.1207 * yselect**3
        a[valid] = 1.752 - 0.316*y - (0.104 / ((y-4.67)**2 + 0.341)) + f_a
        b[valid] = -3.090 + 1.825*y + (1.206 / ((y-4.62)**2 + 0.263)) + f_b

        # far-UV CCM89 extrapolation
        valid = (8 <= x) & (x < 11)
        y = x[valid] - 8.
        coef_a = np.array([-0.070, 0.137, -0.628, -1.073])
        coef_b = np.array([0.374, -0.420, 4.257, 13.670])
        a[valid] = np.polyval(coef_a,y)
        b[valid] = np.polyval(coef_b,y)

        # Overwrite UV with GCC09 model if applicable. Not an extrapolation.
        if model == 'gcc09':
            valid = (3.3 <= x) & (x < 11)
            y = x[valid]
            f_a = np.zeros(y.size)
            f_b = np.zeros(y.size)
            select = (5.9 <= y)
            yselect = y[select] - 5.9
            f_a[select] = -0.110 * yselect**2 - 0.0099 * yselect**3
            f_b[select] = 0.537 * yselect**2 + 0.0530 * yselect**3
            a[valid] = 1.896 - 0.372*y - (0.0108 / ((y-4.57)**2 + 0.0422)) + f_a
            b[valid] = -3.503 + 2.057*y + (0.718 / ((y-4.59)**2 + 0.0530*3.1)) + f_b

        if isinstance(ebv, np.ndarray):
            if (len(ebv.shape) > 1):
                a = a[np.newaxis, np.newaxis, :]; b = b[np.newaxis, np.newaxis, :]
                ebv = ebv[:, :, np.newaxis]
        
        a_v = ebv * r_v
        a_lambda = a_v * (a + b/r_v)
        reddening_curve = 10**(0.4 * a_lambda)

        return reddening_curve

    def median_reddening(self, wavelength, response):
            """
            Compute the mean wavelength given arrays of wavelength (wavelength) and transmission (response).

            Parameters
            ----------
            wavelength : array_like
                Wavelengths (e.g., in Angstroms).
            response : array_like
                Filter transmission values corresponding to wavelength.

            Returns
            -------
            lambda_mean : float
                The mean wavelength.
            """
            # Ensure the inputs are numpy arrays
            wavelength = np.asarray(wavelength)
            response = np.asarray(response) / np.max(np.asarray(response))
            
            # Get reddening curve
            red_curve = self.reddening_ccm(wavelength, ebv=self.ebv, a_v=None, r_v=3.1, model='ccm89')
            A_lambda = 2.5 * np.log10(red_curve)

            # Numerically integrate T(lambda)*lambda over lambda for the numerator
            numerator = np.trapz(response * A_lambda, x=wavelength)
            
            # Numerically integrate response(lambda)/lambda over lambda for the denominator
            denominator = np.trapz(response, x=wavelength)
            
            # Compute the pivot wavelength as the square root of the ratio
            A_mean = numerator / denominator
            return A_mean
        



import math
import numpy as np
from astroquery.ipac.ned import Ned
from astropy.wcs import WCS
from photutils.segmentation import detect_sources
from photutils.background import Background2D, MedianBackground
from photutils.detection import DAOStarFinder
from astropy.stats import sigma_clipped_stats
from photutils.psf import PSFPhotometry, MoffatPSF

from ..utils.utility import useful_functions

class Masking:
    def __init__(self):
        pass
    
    @classmethod
    def adapt_mask(cls, image_data, header, error_data, galaxy_name, observatory, band, image_set):
        fwhm = image_set.psf[galaxy_name][observatory][band]  # in "
        pixel_scale = np.abs(header.get("CD1_1", 1.1e-4)) * 3600
        fwhm = fwhm / pixel_scale  # in pixel
        
        mask_image, masked_image, _ = cls.make_mask(cls, image_data, header, galaxy_name, fwhm)
        masked_err = np.where(mask_image, 999.0, error_data)
        
        image_set.update_data(masked_image, galaxy_name, observatory, band)
        image_set.update_error(masked_err, galaxy_name, observatory, band)
        

    def make_mask(self, image, header, galaxy, psf_fwhm):
        ra, dec = Ned.query_object(galaxy)['RA', 'DEC'][0]

        wcs = WCS(header)
        x, y = wcs.all_world2pix(ra, dec, 0)

        bkg_estimator = MedianBackground()
        try:    
            bkg = Background2D(image, (500, 500), filter_size=(13, 13), bkg_estimator=bkg_estimator)
        except ValueError:
            print('ValueError occured. Try Smaller Background size.')
            bkg = Background2D(image, (100, 100), filter_size=(13, 13), bkg_estimator=bkg_estimator)
        threshold = 1.5*bkg.background_rms

        segment_map = detect_sources(image, threshold, npixels=5)
        if segment_map == None:
            return image, image, image

        sky_map = np.nonzero(segment_map.data)
        sky_image = image.copy()
        sky_image[sky_map] = np.nan

        label_main = segment_map.data[int(y), int(x)]
        if label_main != 0:
            segment_map.remove_labels([label_main])

        mask = np.nonzero(segment_map.data)
        masked_image = image.copy()
        masked_image[mask] = np.nan
            
        mask_image = np.zeros_like(image)
        mask_image[mask] = image[mask]
        
        mean, median, std = sigma_clipped_stats(image, sigma=3.0)
        daofind = DAOStarFinder(fwhm=psf_fwhm, threshold=30.*std)
        sources = daofind(masked_image - mean)

        self.x0, self.y0, self.a, self.b, self.theta = useful_functions.get_galaxy_radius(image)

        dist = self.is_point_in_rotated_ellipse(self, sources['xcentroid'], sources['ycentroid'])

        good = dist > 1
        filtered_sources = sources[good]
        if not filtered_sources.indices:
            return image, image, image

        psf_model = MoffatPSF()
        psf_model.alpha.fixed = False
        psf_model.flux.fixed = False
        fit_shape = (31, 31) 
        psfphot = PSFPhotometry(psf_model, fit_shape,
                                aperture_radius=psf_fwhm*5)
        phot = psfphot(masked_image - mean, init_params=filtered_sources["xcentroid", "ycentroid", "flux"])
        if phot is None:
            print('No PSF photometry found.')
            return masked_image, masked_image, masked_image
        
        resid = psfphot.make_residual_image(masked_image - mean)
        mask_image = np.where(masked_image - mean - resid > 0, 1, 0)
        masked_image = resid.copy()
        
        return mask_image, masked_image, sky_image
        

    def is_point_in_rotated_ellipse(self, x, y):

        dx = x - self.x0
        dy = y - self.y0

        cos_t = math.cos(self.theta)
        sin_t = math.sin(self.theta)
        x_rot =  dx * cos_t + dy * sin_t
        y_rot = -dx * sin_t + dy * cos_t

        value = (x_rot / self.a)**2 + (y_rot / self.b)**2
        return value



import numpy as np
from pathlib import Path
from astropy.io import fits
import pandas as pd
import re
import inspect
import warnings
from glob import glob
from functools import wraps
from typing import Callable, Optional, Union, List

from .utility import Filters, Observatories
from .utility import useful_functions
from .file_generator import *


class GalaxyImageSet():
    __signature__ = inspect.Signature()
    
    def __init__(self):
        self.data = {}
        self.header = {}
        self.error = {}
        self.psf = {}
        self.obs = {}
        self.files = []
        
    def add_image(self, filepath):
        """
        Add an image to the dataset under a galaxy and band identifier.
        Automatically loads the FITS data into memory.
        """
        filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError(f"File {filepath} not found")

        with fits.open(filepath) as hdul:
            image_data = hdul[0].data
            image_header = hdul[0].header

        # Parse galaxy name, band, and observatory from the filename
        file_name = filepath.stem
        galaxy_name = Parsers._galaxy_name_parser(file_name=file_name)
        band = Parsers._band_name_parser(file_name=file_name)
        observatory = Parsers._observatory_name_parser(file_name=file_name)
        
        if galaxy_name is None or band is None or observatory is None:
            return
        
        if galaxy_name not in self.data:
            self.data[galaxy_name] = {}
            self.header[galaxy_name] = {}
            self.error[galaxy_name] = {}
            self.psf[galaxy_name] = {}
            
        if observatory not in self.data[galaxy_name]:
            self.data[galaxy_name][observatory] = {}
            self.header[galaxy_name][observatory] = {}
            self.error[galaxy_name][observatory] = {}
            self.psf[galaxy_name][observatory] = {}
                
        self.data[galaxy_name][observatory][band] = image_data
        self.header[galaxy_name][observatory][band] = image_header
        self.psf[galaxy_name][observatory][band] = -1.0
        self.files.append(str(filepath))
        
        # Find error map and add to variable
        err_name = filepath.with_name(filepath.stem + "_err").with_suffix(filepath.suffix)
        
        if not err_name.exists():
            warnings.warn(f"Error File {err_name} not found. Continue without error map.")
            self.error[galaxy_name][observatory][band] = np.zeros_like(image_data)
        else:
            with fits.open(err_name) as hdul_err:
                error_data = hdul_err[0].data
            
            self.error[galaxy_name][observatory][band] = error_data
        
    # Update image data
    def update_data(self, image_data, galaxy_name, observatory, band):
        """
        Update an image to the dataset under a galaxy and band identifier.
        """
        # print(self.data[galaxy_name][observatory][band].shape)
        # print(np.nansum(self.data[galaxy_name][observatory][band]))
        
        if galaxy_name is None or band is None or observatory is None:
            raise KeyError("Specify the galaxy name, band and observatory name.")
        
        if (galaxy_name not in self.data 
            or observatory not in self.data[galaxy_name] 
            or band not in self.data[galaxy_name][observatory]):
            raise KeyError("Input valid name.")
                
        self.data[galaxy_name][observatory][band] = image_data
        # print(self.data[galaxy_name][observatory][band].shape)
        # print(np.nansum(self.data[galaxy_name][observatory][band]))
        
    def update_error(self, error_data, galaxy_name, observatory, band):
        """
        Update an image to the dataset under a galaxy and band identifier.
        """
        
        if galaxy_name is None or band is None or observatory is None:
            raise KeyError("Specify the galaxy name, band and observatory name.")
        
        if (galaxy_name not in self.error 
            or observatory not in self.error[galaxy_name] 
            or band not in self.error[galaxy_name][observatory]):
            raise KeyError("Input valid name.")
                
        self.error[galaxy_name][observatory][band] = error_data
        
    # merging / append instance
    def append(self, other):
        """
        Append data from another GalaxyImageSet instance to this instance.
        This modifies the current instance.
        
        Args:
            other (GalaxyImageSet): Another GalaxyImageSet instance to append from
        """
        if not isinstance(other, GalaxyImageSet):
            raise TypeError("Can only append another GalaxyImageSet instance")
        
        # Merge data
        for galaxy_name, observatories in other.data.items():
            if galaxy_name not in self.data:
                self.data[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self.data[galaxy_name]:
                    self.data[galaxy_name][observatory] = {}
                
                for band, image_data in bands.items():
                    # Override if exists, add if new
                    self.data[galaxy_name][observatory][band] = image_data
        
        # Merge headers
        for galaxy_name, observatories in other.header.items():
            if galaxy_name not in self.header:
                self.header[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self.header[galaxy_name]:
                    self.header[galaxy_name][observatory] = {}
                
                for band, header_data in bands.items():
                    self.header[galaxy_name][observatory][band] = header_data
        
        # Merge error data
        for galaxy_name, observatories in other.error.items():
            if galaxy_name not in self.error:
                self.error[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self.error[galaxy_name]:
                    self.error[galaxy_name][observatory] = {}
                
                for band, error_data in bands.items():
                    self.error[galaxy_name][observatory][band] = error_data
        
        # Merge PSF data
        for galaxy_name, observatories in other.psf.items():
            if galaxy_name not in self.psf:
                self.psf[galaxy_name] = {}
            
            for observatory, bands in observatories.items():
                if observatory not in self.psf[galaxy_name]:
                    self.psf[galaxy_name][observatory] = {}
                
                for band, psf_data in bands.items():
                    self.psf[galaxy_name][observatory][band] = psf_data
        
        # Merge obs data
        for galaxy_name, obs_data in other.obs.items():
            if galaxy_name not in self.obs:
                self.obs[galaxy_name] = obs_data
            else:
                # If both have obs data for the same galaxy, merge them
                if isinstance(self.obs[galaxy_name], dict) and isinstance(obs_data, dict):
                    self.obs[galaxy_name].update(obs_data)
                else:
                    self.obs[galaxy_name] = obs_data
        
        # Merge file lists (avoid duplicates)
        for filepath in other.files:
            if filepath not in self.files:
                self.files.append(filepath)


    def merge(self, other):
        """
        Create a new GalaxyImageSet instance containing data from both instances.
        This does not modify either original instance.
        
        Args:
            other (GalaxyImageSet): Another GalaxyImageSet instance to merge with
            
        Returns:
            GalaxyImageSet: A new instance containing merged data
        """
        if not isinstance(other, GalaxyImageSet):
            raise TypeError("Can only merge with another GalaxyImageSet instance")
        
        # Create new instance
        merged = GalaxyImageSet()
        
        # First append self to the new instance
        merged.append(self)
        
        # Then append other to the new instance
        merged.append(other)
        
        return merged


    # Get data
    def get_galaxy_data(self, galaxy_name):
        """Return all band images for a specific galaxy."""
        return self.data.get(galaxy_name, {})
    
    def get_observatory_image(self, galaxy_name, observatory):
        """Return image of a specific observatory for a specific galaxy"""
        return self.data.get(galaxy_name, {}).get(observatory, None)
    
    def get_band_image(self, galaxy_name, observatory, band):
        """Return image of a specific band for a specific galaxy."""
        return self.data.get(galaxy_name, {}).get(observatory, {}).get(band, None)
    
    def get_header(self, galaxy_name=None, band=None):
        """Return header information for a specific galaxy or band."""
        return self.header.get(galaxy_name, {}).get(band, None) if galaxy_name else self.header

    def get_psf(self, galaxy_name=None):
        return useful_functions.extract_values_recursive(self.psf, galaxy_name)
    
    
    # Set data
    def set_psf(self, galaxy_name, observatory, band, new_psf_value):
        self.psf[galaxy_name][observatory][band] = new_psf_value

    def summary(self):
        """Print summary of galaxies and available bands."""
        for galaxy, val in self.data.items():
            for obs, bands in val.items():
                galaxy = sorted(galaxy) if isinstance(galaxy, list) else [galaxy]
                obs = sorted(obs) if isinstance(obs, list) else [obs]
                print(f"Galaxy: {galaxy}\nObservatories: {obs} \nBands: {sorted(bands.keys())}")
                
    # Get file from instance
    def get_input_file(self):
        pass
        

class Parsers:
    def __init__(self):
        pass
    
    @staticmethod
    def _observatory_name_parser(file_name):
        """Parse observatory names from a given string or list."""
        if not isinstance(file_name, str):
            raise ValueError("file_name must be a string")
        for obs in Observatories.get_observatories():
            pattern = "|".join(map(re.escape, ['-', ' ', '_', '.']))
            if obs in re.split(pattern, file_name):
                return obs
        return None
    
    @staticmethod
    def _galaxy_name_parser(file_name):
        """Parse galaxy names from a given string or list."""
        if not isinstance(file_name, str):
            raise ValueError("file_name must be a string")
        for galaxy_category in ['NGC', 'IC', 'M']:
            pattern = "|".join(map(re.escape, ['-', ' ', '_', '.']))
            match = [g for g in re.split(pattern, file_name) if galaxy_category in g]
            if match:
                return match[0]
        return None
    
    @staticmethod
    def _band_name_parser(file_name):
        """Parse band names from a given string or list."""
        if not isinstance(file_name, str):
            raise ValueError("file_name must be a string")
        for band in Filters.get_filters():
            pattern = "|".join(map(re.escape, ['-', ' ', '_', '.']))
            if band in re.split(pattern, file_name):
                return band
        return None
    

class ImageQuery():
    def __init__(self):
        pass    
                
    def queryImages(self, dir, galaxy_name=None, band=None):
        """
        Query images from a directory based on galaxy names and bands.
        Returns a GalaxyImageSet object containing the queried images.
        """
        image_set = GalaxyImageSet()
        dir_path = Path(dir).parent
        file_pattern = Path(dir).name if Path(dir).name.endswith('.fits') else "*.fits"
        
        if not dir_path.is_dir():
            raise NotADirectoryError(f"{dir} is not a valid directory")

        if isinstance(galaxy_name, str):
            galaxy_name = [galaxy_name]
        if isinstance(band, str):
            band = [band]

        for fits_file in dir_path.glob(file_pattern):
            try:
                image_set.add_image(fits_file)
            except Exception as e:
                print(f"Error processing {fits_file}: {e}")
                    
        return image_set
    
    @classmethod
    def queryAllImages(cls, dir):
        image_set = GalaxyImageSet()
        
        if not isinstance(dir, str):
            dir = str(dir)
        
        for file in list(glob(dir)):
            if Path(file).stem.endswith("_err"):
                continue
            image_set.add_image(file)
            
        return image_set


import time
import pandas as pd
import numpy as np

from astropy.table import Table

from .utility import Filters, useful_functions

class inputGenerator:
    def __init__(self):
        pass
    
    @classmethod
    def dataframe_generator(cls, image_set, cat_type):
        input_df = pd.DataFrame(columns=['id'])
        
        for i, ((galaxy, obs, band), values) in enumerate(useful_functions.tour_nested_dict_with_keys(image_set.data)):
            # With the first data, make structure of dataframe
            if i == 0:
                x, y = values.shape
                ids = pd.DataFrame([f"{galaxy}_{i}" for i in range(int(x * y))], columns=["id"])
                input_df = pd.concat([input_df, ids])
            else:
                pass
            
            im_series = pd.Series(np.array(values).flatten(), name=f"{obs}.{band}")
            input_df = pd.concat([input_df, im_series], axis=1)
            
            err_series = pd.Series(np.array(image_set.error[galaxy][obs][band]).flatten(), name=f"{obs}.{band}_err")
            input_df = pd.concat([input_df, err_series], axis=1)
            
        # drop NaN values
        float_cols = input_df.select_dtypes(include=['floating']).columns
        input_df[float_cols] = input_df[float_cols].astype('float64')  

            
        input_df.loc[:, input_df.columns != 'id'] = input_df.loc[:, input_df.columns != 'id'].where(input_df.loc[:, input_df.columns != 'id'] >= 0, 0).fillna(0)
        input_df = input_df.astype({'id': 'str'})
        input_df = input_df[(input_df.loc[:, input_df.columns != 'id'] != 0).all(axis=1)]
        
        colnames = Filters.get_catcols(cat_type)
        input_df.rename(columns=colnames, inplace=True)
            
        return input_df

"""input_df = pd.DataFrame(columns=['id']+BANDs)

for id in list(set([i.split('-')[0] for i in region_dic.keys()])):
    ids = pd.DataFrame([f'{id}_{i}' for i in range(bin_size ** 2)], columns=['id'])
    input_df = pd.concat([input_df, ids])
    
for i, name in enumerate(list(region_dic.keys())):
    id = name.split('-')[0]
    band = name.split('-')[2]
    flatten_img = np.array(region_dic[name]).flatten()
    
    input_df.loc[input_df['id'].str.contains(id), band] = flatten_img
    
    if err_bool[name]:
        if f'{band}_err' not in input_df.columns:
            input_df[f'{band}_err'] = ''
        flatten_img = np.array(region_err_dic[name]).flatten()
        input_df.loc[input_df['id'].str.contains(id), f'{band}_err'] = flatten_img
    else:
        input_df.loc[input_df['id'].str.contains(id), f'{band}_err'] = flatten_img * 0.1

input_df = input_df.astype('float64', errors='ignore')
# input_df = input_df.drop(['F657N', 'F658N'], axis=1)

# nan_counts = input_df.isnull().sum(axis=1)
# neg_counts = input_df[input_df.loc[:, (input_df.columns != 'id')] <= 0].sum(axis=1)
# input_df[(nan_counts > 1) | (neg_counts < 0), input_df.columns != 'id'] = 0
input_df.loc[:, input_df.columns != 'id'] = input_df.loc[:, input_df.columns != 'id'].where(input_df.loc[:, input_df.columns != 'id'] >= 0, 0).fillna(0)
input_df = input_df.astype({'id': 'str'})

for _, key in enumerate(redshifting.keys()):
    input_df.loc[input_df['id'].str.contains(key), 'redshift'] = redshifting[key]

print(input_df)
# input_df = input_df.drop(columns = ['m550', 'm575'])
input_df = input_df[(input_df.loc[:, input_df.columns != 'id'] != 0).all(axis=1)]



band_for_galapy = {
    'u': 'SDSS.u',
    'g': 'SDSS.g',
    'r': 'SDSS.r',
    'i': 'SDSS.i',
    'z': 'SDSS.z',
    'ch1': 'Spitzer.IRAC.I1',
    'ch2': 'Spitzer.IRAC.I2',
    'F657N': 'HST.WFC3.UVIS1.F657N',
    'F658N': 'HST.WFC3.UVIS1.F658N',
    'u_err': 'SDSS.u_err',
    'g_err': 'SDSS.g_err',
    'r_err': 'SDSS.r_err',
    'i_err': 'SDSS.i_err',
    'z_err': 'SDSS.z_err',
    'ch1_err': 'Spitzer.IRAC.I1_err',
    'ch2_err': 'Spitzer.IRAC.I2_err',
    'F657N_err': 'HST.WFC3.UVIS1.F657N_err',
    'F658N_err': 'HST.WFC3.UVIS1.F658N_err'
}
band_for_7DT = {f'm{int(wav)}':f'7DT.m{int(wav)}' for wav in np.linspace(400, 875, 20)}
band_for_cigale.update(band_for_7DT)
band_for_cigale.update({f'{key}_err': f'{band_for_cigale[key]}_err' for key in band_for_cigale.keys() if '_err' not in key})

input_df.rename(columns=band_for_cigale, inplace=True)


# sel_id = [f'{obj}_{id * bin_size + 73 + k}' for obj in ['NGC3627', 'NGC4254', 'NGC4303', 'NGC4535'] for id in range(bin_size) for k in range(6)]
# input_df = input_df[input_df.id.isin(sel_id)]
input_df.reset_index()

with pd.option_context('display.max_rows', 500, 'display.max_columns', 20): 
    print(input_df)
tbl = Table.from_pandas(input_df)
date_now = time.strftime('%Y%m%d-%H%M', time.localtime())
tbl.write(f'6_CIGALE_inputs/{date_now}_cigale_table.fits', overwrite=True)

end_time = datetime.datetime.now()
print(f'Start at: {start_time}')
print(f'End at: {end_time}')
print(f'Time taken: {end_time - start_time}')"""


import inspect
import numpy as np

class ImageProcessingPipeline:
    def __init__(self, galaxy_image_set):
        self.galaxy_image_set = galaxy_image_set
        self.pipeline_steps = []
        self.step_configs = {}
    
    def add_step(self, function, config=None, step_name=None):
        """파이프라인에 처리 단계 추가"""
        if step_name is None:
            step_name = function.__name__
        
        self.pipeline_steps.append({
            'name': step_name,
            'function': function,
            'config': config or {}
        })
        
    def remove_step(self, step_name):
        """특정 단계 제거"""
        self.pipeline_steps = [step for step in self.pipeline_steps 
                              if step['name'] != step_name]
    
    def execute(self, galaxy_filter=None, observatory_filter=None, band_filter=None):
        """파이프라인 실행"""
        # 필터 조건에 맞는 이미지들을 선택
        targets = self._get_filtered_targets(galaxy_filter, observatory_filter, band_filter)
        
        for galaxy, observatory, band, image, header, error in targets:
            for step in self.pipeline_steps:
                self._execute_step(step, galaxy, observatory, band, image, header, error)
    
    def _get_filtered_targets(self, galaxy_filter, observatory_filter, band_filter):
        """필터 조건에 맞는 (galaxy, observatory, band) 조합 반환"""
        targets = []
        
        for galaxy in self.galaxy_image_set.data:
            if galaxy_filter and galaxy not in galaxy_filter:
                continue
                
            for observatory in self.galaxy_image_set.data[galaxy]:
                if observatory_filter and observatory not in observatory_filter:
                    continue
                    
                for band in self.galaxy_image_set.data[galaxy][observatory]:
                    if band_filter and band not in band_filter:
                        continue
                    
                    target_image = self.galaxy_image_set.data[galaxy][observatory][band]
                    target_header = self.galaxy_image_set.header[galaxy][observatory][band]
                    try:
                        target_error = self.galaxy_image_set.error[galaxy][observatory][band]
                    except:
                        target_error = np.zeros_like(target_image)
                    
                    targets.append((galaxy, observatory, band,
                                    target_image, target_header, target_error))
        
        return targets
    
    def _execute_step(self, step, galaxy, observatory, band, image, header, error):
        """Execute each step"""
        function = step['function']
        config = step['config']
        
        sig = inspect.signature(function)
        
        # image_data, header, error_data, galaxy_name, observatory, band, image_set
        kwargs = {'image_set': self.galaxy_image_set,
                  'image_data':image, 'header':header, 'error_data':error,
                  'galaxy_name': galaxy, 'observatory': observatory, 'band': band}
        
        kwargs.update(config)
        
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
        
        result = function(**filtered_kwargs)
        print(f"✓ {step['name']} completed for {galaxy}/{observatory}/{band}")
        
        # try:
        #     result = function(**filtered_kwargs)
        #     print(f"✓ {step['name']} completed for {galaxy}/{observatory}/{band}")
        # except Exception as e:
        #     print(f"✗ {step['name']} failed for {galaxy}/{observatory}/{band}: {e}")


from .unit import conversion
from ..reduction.background import backgroundSubtraction
from ..reduction.PSF import PointSpreadFunction
from ..manipulation.reddening import Reddening
from ..manipulation.mask import Masking
from ..manipulation.sky_interpolate import interpolate_sky
from ..division.binning import Bin
from ..division.cutout import CutRegion

from .file_generator import inputGenerator

def execute_pipeline(galaxy_image_set, bin, cat_type):
    pipeline1 = ImageProcessingPipeline(galaxy_image_set)

    pipeline1.add_step(conversion().unitConvertor, step_name="Convert Unit")
    # pipeline1.add_step(backgroundSubtraction)
    pipeline1.add_step(PointSpreadFunction.extract, step_name="Extract PSF")
    
    pipeline1.execute()
    
    pipeline2 = ImageProcessingPipeline(galaxy_image_set)
    
    pipeline2.add_step(PointSpreadFunction.convolution, step_name="Convolve with PSF")
    pipeline2.add_step(Reddening().dered, step_name="Dereddening")
    pipeline2.add_step(Masking.adapt_mask, step_name="Masking")
    pipeline2.add_step(interpolate_sky, step_name="Interpolate Masked Region")
    pipeline2.add_step(Bin.do_binning, config={"bin_size": int(bin)}, step_name="Binning Image")
    
    pipeline2.execute()
    
    pipeline3 = ImageProcessingPipeline(galaxy_image_set)
    
    pipeline3.add_step(CutRegion.cutout_region, config={"box_size" : None}, step_name="Cutout Image Region")
    
    pipeline3.execute()
    
    input_df = inputGenerator.dataframe_generator(galaxy_image_set, cat_type)
    return input_df



from astropy.convolution import convolve, Gaussian2DKernel
import numpy as np

def interpolate_sky(image_data, galaxy_name, observatory, band, image_set):
    # gauss_kernal = Gaussian2DKernel(x_stddev=1)
    # interp_img = convolve(image_data, gauss_kernal, normalize_kernel=True, nan_treatment='interpolate')
    interp_img = np.where(image_data == np.nan, 0, image_data)
    
    image_set.update_data(interp_img, galaxy_name, observatory, band)


from astropy.stats import SigmaClip
from photutils.background import Background2D, MedianBackground

def backgroundSubtraction(image_set, image_data, galaxy_name, observatory, band):
    im_x, im_y = image_data.shape
    box_size = (int(im_x * 0.1), int(im_y * 0.1))
    filter_size = (int(im_x * 5e-3) * 2 + 1, int(im_y * 5e-3) * 2 + 1)
    
    sigma_clip = SigmaClip(sigma=3.0)
    bkg_estimator = MedianBackground()
    bkg = Background2D(image_data, box_size=box_size, filter_size=filter_size,
                       sigma_clip=sigma_clip, bkg_estimator=bkg_estimator)
    bkg_map = bkg.background
    
    image_set.update_data(image_data - bkg_map, galaxy_name, observatory, band)
    


import numpy as np
import math
from photutils.segmentation import detect_threshold, detect_sources, SourceCatalog

class Filters:
    """Class to handle different photometric filters and their properties."""
    def __init__(self):
        self.broadband = self._broadband()
        self.mediumband = self._mediumband()
        self.filters = self.broadband + self.mediumband
    
    def _broadband(self):
        """Return a list of broadband filters."""
        return [
            'FUV', 'NUV', 'u', 'g', 'r', 'i', 'z', 'Y', 'J', 'H', 'Ks',
            'w1', 'w2', 'w3', 'w4'
        ]
    
    def _mediumband(self):
        """Return a list of mediumband filters."""
        return [f'm{wave}' for wave in range(400, 900, 25)]
    
    @classmethod
    def get_filters(cls):
        """Return a list of all filters."""
        return cls().filters
    
    @classmethod
    def get_catcols(cls, cat_type):
        """Return a dictionary of given type"""
        catcols ={"cigale": cls.cigale(cls),
                  "eazy": cls.eazy(cls),
                  "lephare": cls.lephare(cls),
                  "ppxf": cls.ppxf(cls),
                  "goyangyi": cls.goyangyi(cls)
                  }
        
        return catcols[cat_type.lower()]
    
    
    def cigale(self):
        cols_cigale = {
            'GALEX.NUV': 'galex.NUV',
            'GALEX.FUV': 'galex.FUV',
            'SDSS.u': 'SDSS_u',
            'SDSS.g': 'SDSS_g',
            'SDSS.r': 'SDSS_r',
            'SDSS.i': 'SDSS_i',
            'SDSS.z': 'SDSS_z',
            'PanStarrs.y': 'PAN-STARRS_y',
            '2MASS.J': 'J_2mass',
            '2MASS.H': 'H_2mass',
            '2MASS.Ks': 'Ks_2mass',
            'Spitzer.ch1': 'spitzer.irac.ch1',
            'Spitzer.ch2': 'spitzer.irac.ch2',
            'WISE.w1': 'WISE1',
            'WISE.w2': 'WISE2',
            'WISE.w3': 'WISE3',
            'WISE.w4': 'WISE4',
            'F657N': 'HST.UVIS1.F657N',
            'F658N': 'HST.UVIS1.F658N',
        }
        cols_cigale.update({f'{key}_err': f'{cols_cigale[key]}_err' for key in cols_cigale.keys() if '_err' not in key})
        return cols_cigale
    
    def eazy(self):
        cols_eazy = {
            
        }
        return cols_eazy
    
    def lephare(self):
        cols_lephare = {
            
        }
        return cols_lephare
    
    def ppxf(self):
        cols_ppxf = {
            
        }
        return cols_ppxf
    
    
    def goyangyi(self):
        print(" ╱|、\n(˚ˎ 。7  \n |、˜〵          \n じしˍ,)ノ")
        return self.cigale(self)

class Observatories:
    """Class to handle different observatories and their properties."""
    def __init__(self):
        self.optical_obs = self._opticals()
        self.ir_obs = self._infrareds()
        self.uv_obs = self._ultraviolet()
        self.radio_obs = self._radio()
        self.observatories = list(set(self.optical_obs + self.ir_obs + self.uv_obs + self.radio_obs))
        
    def _opticals(self):
        """Return a list of optical observatories."""
        return ['HST', 'SDSS', 'PS1', 'CFHT', 'DECam', 'DES', 'LSST', 'Pan-STARRS', 'Subaru', '7DT', 'SkyMapper']
    
    def _infrareds(self):
        """Return a list of infrared observatories."""
        return ['WISE', 'Spitzer', 'Herschel', 'JWST', 'VISTA', 'UKIDSS', '2MASS']
    
    def _ultraviolet(self):
        """Return a list of ultraviolet observatories."""
        return ['GALEX', 'HST', 'FUSE']
    
    def _radio(self):
        """Return a list of radio observatories."""
        return ['VLA', 'ALMA', 'LOFAR', 'SKA', 'MeerKAT', 'GMRT']
    
    @classmethod
    def get_observatories(cls):
        """Return a list of all observatories."""
        return cls().observatories

class useful_functions:
    
    @classmethod
    def get_galaxy_radius(cls, image):
        
        threshold = detect_threshold(image, nsigma=3.0)

        segm = detect_sources(image, threshold, npixels=5)

        catalog = SourceCatalog(image, segm)
        gal = max(catalog, key=lambda src: src.area)

        x0, y0 = gal.xcentroid, gal.ycentroid
        a, b = gal.semimajor_sigma.value*2, gal.semiminor_sigma.value*2
        theta = math.radians(gal.orientation.value)
        return x0, y0, a, b, theta
    
    @staticmethod
    def extract_values_recursive(dictionary, key):
        """
        Alternative recursive approach that handles arbitrary nesting depth.
        
        Args:
            dictionary: Dictionary with nested structure
            key: The key at level1 to extract values from
        
        Returns:
            List of all values found in the nested structure
        """
        def _extract_all_values(obj):
            """Recursively extract all values from nested dict/list structures."""
            if isinstance(obj, dict):
                values = []
                for v in obj.values():
                    values.extend(_extract_all_values(v))
                return values
            elif isinstance(obj, list):
                values = []
                for item in obj:
                    values.extend(_extract_all_values(item))
                return values
            else:
                return [obj]
        
        if key not in dictionary:
            return []
        
        return _extract_all_values(dictionary[key])
    
    @classmethod
    def tour_nested_dict_with_keys(cls, dictionary):
        """
        Tour through a 3-level nested dictionary and yield keys and values in order.
        
        Args:
            dictionary: Dictionary with structure dict[level1][level2][level3]
        
        Yields:
            tuple: (keys_tuple, value) where keys_tuple contains (level1_key, level2_key, level3_key)
        """
        for level1_key, level1_dict in dictionary.items():
            for level2_key, level2_dict in level1_dict.items():
                for level3_key, value in level2_dict.items():
                    yield (level1_key, level2_key, level3_key), value


    def get_all_keys_and_values(self, my_dict):
        """
        Get all keys and values from a 3-level nested dictionary as a list.
        
        Args:
            my_dict: Dictionary with structure dict[level1][level2][level3]
        
        Returns:
            list: List of tuples [(keys_tuple, value), ...] where keys_tuple contains (level1_key, level2_key, level3_key)
        """
        result = []
        for level1_key, level1_dict in my_dict.items():
            for level2_key, level2_dict in level1_dict.items():
                for level3_key, value in level2_dict.items():
                    result.append(((level1_key, level2_key, level3_key), value))
        return result


    def tour_nested_dict_recursive(self, obj, current_keys=()):
        """
        Recursive function to tour through arbitrarily nested dictionaries.
        
        Args:
            obj: Dictionary or value to traverse
            current_keys: Current key path (used internally)
        
        Yields:
            tuple: (keys_tuple, value) where keys_tuple contains all keys in the path
        """
        if isinstance(obj, dict):
            for key, value in obj.items():
                yield from self.tour_nested_dict_recursive(value, current_keys + (key,))
        else:
            yield current_keys, obj


import numpy as np

from ..utils.utility import useful_functions

class CutRegion:
    
    @classmethod
    def cutout_region(cls, box_size, image_data, error_data, galaxy_name, observatory, band, image_set):
        if box_size == None or not box_size:
            x, y, a, b, th = useful_functions.get_galaxy_radius(image_data)
            a = 1.5 * a; b = 1.5 * b
            box_size = (x, y, a, b, th)
        
        cut_img, cut_error = cls.get_cutout(image_data, error_data, box_size, 'ellipse')
        image_set.update_data(cut_img, galaxy_name, observatory, band)
        image_set.update_error(cut_error, galaxy_name, observatory, band)

    def get_cutout(img, error, size, _shape: str='box'):
        """
        Mask out everything except a central region of shape 'box', 'circle', or 'ellipse'.

        Parameters
        ----------
        img : 2D ndarray
            Input image.
        size : float or tuple of floats
            - If _shape in {'box','circle'}: scalar = side‐length (box) or diameter (circle).
            - If _shape=='ellipse': either
                * scalar = major and minor axes
                * tuple (width, height) in pixels for major/minor axes.
        _shape : {'box','circle','ellipse'}
            Shape of the kept region.

        Returns
        -------
        cutout : 2D ndarray
            Same shape as `img`, with pixels **outside** the requested region set to zero.
        """
        x0, y0, a, b, theta = size
        

        if _shape == 'box':
            from photutils.aperture import RectangularAperture
            mask = RectangularAperture((x0, y0), w=a, h=b, theta=theta)
            mask_im = mask.to_mask().to_image(shape=img.shape)
            
            return img * mask_im.astype(img.dtype), error * mask_im.astype(error.dtype)

        elif _shape == 'circle':
            from photutils.aperture import CircularAperture
            mask = CircularAperture((x0, y0), r=a)
            mask_im = mask.to_mask().to_image(shape=img.shape)
            
            return img * mask_im.astype(img.dtype), error * mask_im.astype(error.dtype)

        elif _shape == 'ellipse':
            from photutils.aperture import EllipticalAperture
            mask = EllipticalAperture((x0, y0), a=a, b=b, theta=theta)
            mask_im = mask.to_mask().to_image(shape=img.shape)
            
            return img * mask_im.astype(img.dtype), error * mask_im.astype(error.dtype)

        else:
            raise ValueError(f"Unknown shape '{_shape}'. Choose 'box', 'circle', or 'ellipse'.")



import numpy as np
from scipy.stats import mode

from photutils.detection import DAOStarFinder
from astropy.stats import sigma_clipped_stats
from astropy.modeling import models, fitting
from astropy.convolution import convolve_fft, Gaussian2DKernel

import warnings
from astropy.utils.exceptions import AstropyWarning
warnings.simplefilter('ignore', category=AstropyWarning)

class PointSpreadFunction:
    
    def __call__(self):
        pass
    
    @classmethod
    def extract(cls, image_data, header, galaxy_name, observatory, band, image_set):    
        fwhm_val = cls.measure_psf_fwhm(cls, image_data, header, threshold_sigma=15)
        
        image_set.set_psf(galaxy_name, observatory, band, fwhm_val) # in " fwhm * pixel_scale
    
    @classmethod
    def convolution(cls, image_data, header, error_data, galaxy_name, observatory, band, image_set):
        """
        Convolve `image` with a Gaussian kernel of width `sigma_extra_pix` (pixels).
        If sigma_extra_pix==0, return original image.
        """
        pixel_scale = np.abs(header.get("CD1_1", 1.1e-4)) * 3600  # in "
        
        psf_list = image_set.get_psf(galaxy_name)
        sig_i = image_set.psf[galaxy_name][observatory][band]
        sig_t = np.max(psf_list)
        sigma_extra = np.sqrt(sig_t**2 - sig_i**2) / pixel_scale
        
        if sigma_extra <= 0:
            return image_data.copy()
        kernel = Gaussian2DKernel(x_stddev=sigma_extra)
        convolved_img = convolve_fft(
            image_data, kernel,
            normalize_kernel=True,
            nan_treatment='interpolate'
        )
        convolved_err = convolve_fft(
            error_data, kernel,
            normalize_kernel=True,
            nan_treatment='interpolate'
        )
        image_set.update_data(convolved_img, galaxy_name, observatory, band)
        image_set.update_error(convolved_err, galaxy_name, observatory, band)
    

    def measure_fwhm_gaussian(image, x_center, y_center, box_size=21):
        """
        Measure FWHM by fitting 2D Gaussian to PSF
        """
        # Extract cutout around star
        y, x = np.ogrid[:box_size, :box_size]
        x_start = int(x_center - box_size//2)
        y_start = int(y_center - box_size//2)
        
        cutout = image[y_start:y_start+box_size, x_start:x_start+box_size]
        
        # Create coordinate grids
        y_grid, x_grid = np.mgrid[:box_size, :box_size]
        
        # Initial parameter guess
        amplitude = np.max(cutout)
        x_mean = box_size // 2
        y_mean = box_size // 2
        
        # Fit 2D Gaussian
        g_init = models.Gaussian2D(amplitude=amplitude, 
                                x_mean=x_mean, y_mean=y_mean,
                                x_stddev=box_size * 0.05, y_stddev=box_size * 0.05)
        fit_g = fitting.TRFLSQFitter()
        g = fit_g(g_init, x_grid, y_grid, cutout)
        
        # Convert stddev to FWHM
        fwhm_x = 2.355 * g.x_stddev.value
        fwhm_y = 2.355 * g.y_stddev.value
        fwhm_avg = (fwhm_x + fwhm_y) / 2
        
        return fwhm_avg

    def measure_psf_fwhm(self, image, header, threshold_sigma=15):
        """
        Complete pipeline: detect stars and measure FWHM
        """
        im_x, im_y = image.shape
        box_size = int((im_x + im_y) * 0.1 / 2)
        mean, median, std = sigma_clipped_stats(image, sigma=10.0)
        
        # Star detection
        daofind = DAOStarFinder(fwhm=3.0, threshold=threshold_sigma*std)
        sources = daofind(image)
        
        if sources is None:
            return -1.0
        
        # Measure FWHM for each detected star
        fwhm_measurements = []
        
        for source in sources:
            x, y = source['xcentroid'], source['ycentroid']
            
            margin = 0.1
            # Skip stars too close to edges
            if (x > im_x * margin and x < im_x * (1 - margin) and 
                y > im_y * margin and y < im_y * (1 - margin)):
                
                try:
                    fwhm = self.measure_fwhm_gaussian(image, x, y, box_size=box_size)
                    if not np.isnan(fwhm) and fwhm > 0:
                        fwhm_measurements.append(fwhm)
                except:
                    continue
        
        fwhm, _ = mode(fwhm_measurements, nan_policy='omit')
        pixel_scale = np.abs(header.get("CD1_1", 1.1e-4)) * 3600  # in "
        
        return fwhm * pixel_scale



import numpy as np
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning) 

class conversion:
    def __init__(self):
        self.obs_match = {'GALEX': self.GALEX,
                         'SDSS': self.SDSS,
                         '7DT': self.SDT,
                         'PS1': self.PanStarr1,
                         'Spitzer': self.Spitzer,
                         'WISE': self.WISE,
                         '2MASS': self.TMASS
                         }
    
    def unitConvertor(self, image_data, header, error_data, galaxy_name, observatory, band, image_set):
        """
        Convert image data to consistent flux units based on observatory and band.
        
        Parameters:
        -----------
        image_data : array-like
            Raw image data to be converted
        header : dict
            FITS header or dictionary containing metadata
        galaxy_name : str
            Name of the galaxy (for logging/tracking purposes)
        observatory : str
            Observatory name (key in obs_match)
        band : str
            Filter/band name
            
        Returns:
        --------
        array-like
            Converted flux data in mJy
        """
        
        # Check if observatory is supported
        if observatory not in self.obs_match:
            raise ValueError(f"Observatory '{observatory}' not supported. "
                           f"Available observatories: {list(self.obs_match.keys())}")
        
        # Get the conversion function
        conversion_func = self.obs_match[observatory]
        
        try:
            # All functions now have standardized signature
            convert_im, convert_err = conversion_func(image_data, header, band), conversion_func(error_data, header, band)
            image_set.update_data(convert_im, galaxy_name, observatory, band)
            image_set.update_error(convert_err, galaxy_name, observatory, band)
                
        except Exception as e:
            print(f"Error converting data for {galaxy_name} from {observatory} {band}: {str(e)}")
            raise
    
    # UVs
    def GALEX(self, image_data, header=None, band=None):
        return np.array(image_data) * 1e-3
    
    
    # Opticals
    def SDSS(self, image_data, header=None, band=None):  # nmgy
        return np.array(image_data) * 3.631 * 10 ** (-3)
    
    def SDT(self, image_data, header, band=None):  # ADU
        try:
            ZP = header['ZP_AUTO']
        except KeyError:
            raise KeyError("Zero Point value is not found.")
        flux = 3631 * (image_data) * 10 ** (-ZP / 2.5) * 1e3
        return flux

    def PanStarr1(self, image_data, header, band=None):
        exptime = header['EXPTIME']
        mag = -2.5 * np.log10(image_data) + 25 + 2.5 * np.log10(exptime)
        flux = 10 ** (-0.4 * mag) * 3631 * 1e3
        return flux
    
    
    # IRs
    def Spitzer(self, image_data, header=None, band=None):  # MJy/sr
        return 1e9 * ((1.1e-4 * 3600 / 206265.0) ** 2) * image_data
    
    def WISE(self, image_data, header=None, band=None):  #ADU/DN
        # WISE flux calibration factors
        F_nu0 = {'w1': 309.540,
                'w2': 171.787,
                'w3': 31.674,
                'w4': 8.363
                }
        M_inst0 = {'w1': 20.5,
                'w2': 19.5,
                'w3': 18.0,
                'w4': 13.0
                }

        image_data = np.where(image_data == 0, 0.1, image_data)
        image_data = np.where(np.isnan(image_data), 0.1, image_data)
        M_cal = M_inst0[band] - 2.5 * np.log10(image_data)
        flux = np.power(10, - M_cal / 2.5) * F_nu0[band] * 1e3
        return flux
    
    def TMASS(self, image_data, header, band=None):
        zp_2mass = {'j': 21.1258,
                    'h': 20.7288,
                    'k': 20.1106
                    }
        zpflux_2mass = {'j': 1594,
                        'h': 1024,
                        'k': 666.7
                        }

        mag = -2.5 * np.log10(image_data) + zp_2mass[header['FILTER']]
        f_nu = 10 ** (-0.4 * mag) * zpflux_2mass[header['FILTER']] * 1e3  # [mJy]
        return f_nu


class Bin:
    def __init__(self):
        pass
    
    @classmethod
    def do_binning(cls, bin_size, image_data, error_data, galaxy_name, observatory, band, image_set):
        # if bin_size is None:
        #     bin_size = image_data.shape[0]
        binned_img = cls.binning(image_data, bin_size, bin_size)
        binned_err = cls.binning_err(error_data, bin_size, bin_size)
        image_set.update_data(binned_img, galaxy_name, observatory, band)
        image_set.update_error(binned_err, galaxy_name, observatory, band)
        
    def binning(image, bin_x, bin_y):
        return image.reshape(bin_x, image.shape[0] // bin_x, bin_y, image.shape[1] // bin_y).sum(3).sum(1)

    def binning_err(image, bin_x, bin_y):
        image = image ** 2
        image = image.reshape(bin_x, image.shape[0] // bin_x, bin_y, image.shape[1] // bin_y).sum(3).sum(1)
        return image ** (0.5)


from setuptools import setup, find_packages

setup(
    name='Spec7DT',
    version='0.9.9',
    description='Spectral image handling package for 7-Dimensional Telescope users by Won-Hyeong Lee',
    author='Won-Hyeong Lee',
    author_email='wohy1220@gmail.com',
    url='https://github.com/Yicircle/Spec7DT',
    install_requires=[
        'numpy',
        'astropy',
        'matplotlib',
        'seaborn',
        'pathlib',
        'photutils',
        'reproject',
    ],
    packages=find_packages(where='src'),
    include_package_data=True,
    package_data={
        "Spec7DT": ["reference/*"],
    },
    package_dir={'': 'src'},
    keywords=[''],
    python_requires='>=3.10',
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],
)